@@ -1,8 +1,10 @@
-from PySide2.QtWidgets import QWidget, QPushButton, QLabel, QLineEdit, QApplication, QComboBox, QGridLayout
+from PySide2.QtWidgets import QWidget, QPushButton, QLabel, QLineEdit, QApplication, QComboBox, QGridLayout, QMessageBox
 from PySide2.QtWidgets import QGraphicsScene, QGraphicsView, QGraphicsPixmapItem, QGraphicsRectItem
 from PySide2.QtGui import QPixmap, QBrush, QColor
-from PySide2.QtCore import Qt, QTimer
-from urllib.request import urlopen
+from PySide2.QtCore import Qt, QTimer, QEvent
+from urllib.request import build_opener
+from simpleaudio import WaveObject
+from random import randint
 
 class SimGraphicsView(QGraphicsView):
   def __init__(self, scene, key_handler):
@@ -11,6 +13,13 @@ class SimGraphicsView(QGraphicsView):
   def keyPressEvent(self, event):
     self.key_handler(event)
 
+class MyWidget(QWidget):
+  def __init__(self, key_handler):
+      super().__init__()
+      self.key_handler=key_handler
+  def keyPressEvent(self, event):
+    self.key_handler(event)
+
 class SimGuiApp(QApplication):
     SCENE_WIDTH=400
     SCENE_HEIGHT=300
@@ -19,6 +28,7 @@ class SimGuiApp(QApplication):
         self.mod=None
     def start(self, mod):
         self.mod=mod
+        self.in_modal=False
         self.key_ev=None
         self.gs=None
         self.gv=None
@@ -28,17 +38,30 @@ class SimGuiApp(QApplication):
         self.last_row=None
         self.auto_row=0
         self.auto_col=0
-        wid=QWidget()
-        wid.setWindowTitle("simgui")
+        self.make_opener()
+        self.wid=MyWidget(self.on_key)
+        self.wid.setWindowTitle("simgui")
         self.lo=QGridLayout()
-        wid.setLayout(self.lo)
-        self.call_handler("on_ready")
-        wid.show()
+        self.wid.setLayout(self.lo)
+        self.wid.show()
+        ev=QEvent(QEvent.Type.User)
+        self.postEvent(self, ev)
         self.exec_()
-    def call_handler(self, fn):
+    def event(self, ev):
+      if ev.type()==QEvent.Type.User:
+        self.call_handler("on_ready")  
+      return super().event(ev)
+    def make_opener(self):
+      self.op=build_opener()
+      self.op.addheaders=[("User-agent", "Mozilla/5.0 (X11; U; Linux i686) Gecko/20071127 Firefox/2.0.0.11")]
+      self.op.cache={}
+    def call_handler(self, fn, data=None):
         handler=self.mod.get(fn)
         if handler:
-          handler()
+          if data:
+            handler(data)
+          else:
+            handler()
     def add_label(self, name, text, **kwargs):
         lbl=QLabel(str(text))
         self.add_wid(name, lbl, **kwargs)
@@ -56,8 +79,15 @@ class SimGuiApp(QApplication):
     def set_wid_color(self, name, color):
       wid=self.get_wid(name)
       wid.setStyleSheet(f"background-color: {color}")
+    def fetch_web_data(self, url):
+      if url in self.op.cache:
+        return self.op.cache[url]
+      else:
+        d=self.op.open(url).read()
+        self.op.cache[url]=d
+        return d
     def set_label_img(self, name, img_url):
-      data=urlopen(img_url).read()
+      data=self.fetch_web_data(img_url)
       pm=QPixmap()
       pm.loadFromData(data)
       lbl=self.get_wid(name)
@@ -119,27 +149,37 @@ class SimGuiApp(QApplication):
         return t
     def set_input_text(self, name, text):
       self.get_wid(name).setText(str(text))
+    def on_key(self, event):
+      self.key_ev=event
+      self.call_handler("on_key")      
     def add_graphics_view(self, min_w, min_h):
         if self.gs:
           raise ValueError("Only one graphics view can be added")
-        def on_key(event):
-          self.key_ev=event
-          self.call_handler("on_key")
         self.gs=QGraphicsScene()
-        self.gv=SimGraphicsView(self.gs, on_key)
+        self.gv=SimGraphicsView(self.gs, self.on_key)
         self.gv.setMinimumSize(min_w, min_h)
-        self.gv.setSceneRect(0, 0, SimGuiApp.SCENE_WIDTH, SimGuiApp.SCENE_HEIGHT)
+        #there is a 1 pixel margin hard coded
+        self.gv.setSceneRect(0, 0, SimGuiApp.SCENE_WIDTH-2, SimGuiApp.SCENE_HEIGHT-2)
         self.add_wid("simgui_gv", self.gv)
-    def add_gi_img(self, name, x, y, w, h, img_url):
-      data=urlopen(img_url).read()
-      pm=QPixmap()
-      pm.loadFromData(data)
+    def add_gi_img(self, name, x, y, w, h, img_url_or_file):
+      if img_url_or_file.find("://")>0:
+        data=self.fetch_web_data(img_url_or_file)
+        pm=QPixmap()
+        pm.loadFromData(data)
+      else:
+        from pathlib import Path
+        p=Path(img_url_or_file)
+        if p.exists():
+          pm=QPixmap(img_url_or_file)
+        else:
+          raise ValueError(f"File {img_url_or_file} not found")
       pm2=pm.scaled(w, h, Qt.AspectRatioMode.IgnoreAspectRatio, Qt.TransformationMode.SmoothTransformation)
       gi=QGraphicsPixmapItem(pm2)
       gi.setPos(x, y)
       self.add_gi(name, gi)
     def add_gi_rect(self, name, x, y, w, h, color):
-      gi=QGraphicsRectItem(x, y, w, h)
+      gi=QGraphicsRectItem(0, 0, w, h)
+      gi.setPos(x, y)
       br=QBrush(QColor(color))
       gi.setBrush(br)
       self.add_gi(name, gi)
@@ -159,11 +199,24 @@ class SimGuiApp(QApplication):
     def get_gi_y(self, name):
       gi=self.get_gi(name)
       return gi.pos().y()
+    def gi_exists(self, name):
+      return  name in self.gi_dict
     def get_gi(self, name):
       if name in self.gi_dict:
         return self.gi_dict[name]
       else:
         raise ValueError(f"No graphics item named {name}")
+    def remove_gi(self, name):
+      if self.gs==None:
+        raise ValueError("Must add a graphics scene first")
+      gi=self.get_gi(name)
+      del self.gi_dict[name]
+      self.gs.removeItem(gi)
+    def make_unique_name(self, prefix):
+      while True:
+        name=prefix+str(randint(0, 65535))
+        if not(name in self.gi_dict) and not(name in self.wid_dict):
+          return name
 
     def get_key(self):
       code_map={Qt.Key_Left: "Left", Qt.Key_Right: "Right", Qt.Key_Up: "Up", Qt.Key_Down: "Down", \
@@ -180,13 +233,38 @@ class SimGuiApp(QApplication):
         return "Unknown"
     def start_timer(self, name, interval):
       def on_timeout():
-        self.call_handler("on_timeout_"+name)
+        if not self.in_modal:
+          self.call_handler("on_timeout_"+name)
       tm=QTimer()
       tm.timeout.connect(on_timeout)
       tm.start(int(interval*1000))
       self.timer_dict[name]=tm
     def stop_timer(self, name):
       self.timer_dict[name].stop()
+    def send_data_to_future(self, data, interval):
+      def on_timeout():
+        tm.stop()
+        self.call_handler("on_data_from_past", data)
+      tm=QTimer()
+      tm.timeout.connect(on_timeout)
+      tm.start(int(interval*1000))
+    def are_gi_overlap(self, n1, n2):
+      gi1=self.get_gi(n1)
+      gi2=self.get_gi(n2)
+      r1=gi1.boundingRect()
+      r1=gi1.mapRectToParent(r1)
+      r2=gi2.boundingRect()
+      r2=gi2.mapRectToParent(r2)
+      return r1.intersects(r2)
+    def msg_box(self, text):
+      self.in_modal=True
+      try:
+        QMessageBox.information(self.wid, "Info", str(text))
+      finally:
+        self.in_modal=False
+    def play_wav(self, path):
+      wo=WaveObject.from_wave_file(path)
+      wo.play()
 
 sgapp=SimGuiApp()
 
@@ -244,6 +322,9 @@ def add_gi_img(name, x, y, w, h, img_url):
 def add_gi_rect(name, x, y, w, h, color):
   sgapp.add_gi_rect(name, x, y, w, h, color)
 
+def remove_gi(name):
+  sgapp.remove_gi(name)
+
 def get_key():
   return sgapp.get_key()
 
@@ -261,3 +342,24 @@ def start_timer(name, interval):
 
 def stop_timer(name):
   sgapp.stop_timer(name)
+
+def make_unique_name(prefix):
+  return sgapp.make_unique_name(prefix)
+
+def gi_exists(name):
+  return sgapp.gi_exists(name)
+
+def are_gi_overlap(n1, n2):
+  return sgapp.are_gi_overlap(n1, n2)
+
+def msg_box(text):
+  sgapp.msg_box(text)
+
+def quit():
+  sgapp.quit()
+
+def play_wav(path):
+  sgapp.play_wav(path)
+
+def send_data_to_future(data, interval):
+  sgapp.send_data_to_future(data, interval)
\ No newline at end of file
